"""a modified pidgy kernel.

the command line application installs the kernel"""

from contextlib import suppress

from ipykernel.ipkernel import IPythonKernel


class Kernel(IPythonKernel):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        from . import load_ipython_extension

        load_ipython_extension(self.shell)
        with suppress(BaseException):
            from .widgets import load_ipython_extension

            load_ipython_extension(self.shell)


if __name__ == "__main__":
    import ipykernel.kernelapp

    ipykernel.kernelapp.IPKernelApp.launch_instance(kernel_class=Kernel)



def load_ipython_extension(shell):
    from .weave import IPyWidgetsHtml

    shell.displays_manager.template_cls = IPyWidgetsHtml

    import ipywidgets
    from ipywidgets import Widget

    for k, v in vars(ipywidgets).items():
        if isinstance(v, type) and issubclass(v, Widget):
            if k[0].isupper():
                shell.user_ns.setdefault(k, v)


def unload_ipython_extension(_):
    pass



from collections import defaultdict
from dataclasses import dataclass, field
from io import StringIO
from re import compile
from typing import ChainMap

from jinja2 import Environment, Template
from jinja2.meta import find_undeclared_variables
from markdown_it import MarkdownIt
from traitlets import Bool, Dict, Instance, Type

from .pidgy import Extension

CELL_MAGIC = compile("^\s*%{2}\S")
NO_SHOW = compile(r"^\s*\r?\n")

from IPython import get_ipython


class Finalizer:
    def __new__(cls, object):
        return object


class IPythonFinalizer(Finalizer):
    @staticmethod
    def normalize(type, object, metadata) -> str:
        """normalize and object with (mime)type and return a string."""

        if type == "text/html" or "svg" in type:
            object = get_minified(object)

        if type.startswith("image"):
            md = metadata.get(type, {})
            width, height = md.get("width"), md.get("height")
            object = get_decoded(object)
            *_, data = type.partition("/")
            object = f"""<img src="data:image/{data};base64,{object}"/>"""

        return object

    def __new__(cls, object):
        """convert an object into a markdown/html representation"""
        from IPython import get_ipython

        shell = get_ipython()
        datum = shell.display_formatter.format(object)
        data, metadata = datum if isinstance(datum, tuple) else (datum, {})
        key = next(filter(data.__contains__, get_active_types(shell)), str(object))
        if key == "text/plain":
            return str(object)
        return cls.normalize(key, data[key], metadata)


class IPythonEnvironment(Environment):
    def init_filters(self):
        try:
            from nbconvert.exporters.templateexporter import default_filters

            self.filters.update(default_filters)
        except ModuleNotFoundError:
            pass

    def __init__(self, *args, **kwargs):
        from jinja2 import ChoiceLoader, DictLoader, Environment, FileSystemLoader

        kwargs.setdefault("loader", ChoiceLoader([DictLoader({}), FileSystemLoader(".")]))
        kwargs.setdefault("finalize", IPythonFinalizer)
        kwargs.setdefault("undefined", Undefined)
        kwargs.setdefault("enable_async", False)  # enable this later
        super().__init__(*args, **kwargs)
        self.init_filters()


from jinja2 import Undefined


class IPythonTemplate(Template):
    def ns(self, *args, **kwargs):
        import builtins

        ns = get_ipython()
        if ns:
            return ChainMap(kwargs, ns.user_ns, vars(builtins))
        return {}

    def render(self, *args, **kwargs):
        return super().render(self.ns(*args, **kwargs))

    async def render_async(self, *args, **kwargs):
        return await super().render_async(self.ns(*args, **kwargs))


class Undefined(Undefined):
    def _fail_with_undefined_error(self, *args, **kwargs):
        # log that the template failed
        return f"`{self._undefined_name} is undefined`"


@dataclass
class TemplateDisplay:
    widget = False
    from IPython.display import Markdown

    template: object = None
    display_cls: type = Markdown
    display_handle: object = None
    iframe_attrs: dict = field(default_factory=dict(width="100%", height=600, loading="lazy").copy)
    is_list_urls: bool = None
    vars: set = None
    # this is not used by the base class but may be used by other base classes that render html
    markdown_renderer: object = field(default_factory=MarkdownIt)

    del Markdown

    def _ipython_display_(self):
        self.display()

    def _is_list_urls(self, x):
        for line in filter(bool, map(str.strip, StringIO(x))):
            if "://" in line:
                continue
            return False
        return True

    def render(self):
        render = self.template.render()
        if self.is_list_urls is None:
            self.is_list_urls = self._is_list_urls(render)

        # if self.is_list_urls:
        #     render = self.embed(render)

        return render

    def update(self, change=None):
        # it should be possible to do smarter updates
        if self.display_handle:
            if self.is_widget():
                self.display_handle.value = self.render()
            else:
                self.display_handle.update(self.display_cls(self.render()))

    def embed(self, urls):
        """we have a feature for showing iframes of urls.

        we can add richer features later like domain rules and file extension dispatchers
        maybe the could have been done with markdown it?"""
        lines = []

        # compose the default iframe attributes
        args = " ".join(f'{k}="{v}"' for k, v in self.iframe_attrs.items())

        # iterate through all the lines of the source and generate iframes
        # from them.
        for line in filter(bool, map(str.strip, StringIO(urls))):
            lines.append(f'<iframe src="{line}" {args}/>')
        return "\n".join(lines)


@dataclass
class IPythonMarkdown(TemplateDisplay):
    def display(self):
        from IPython.display import display

        if self.display_handle is None:
            from IPython.display import DisplayHandle

            self.display_handle = DisplayHandle()

        object = self.display_cls(self.template.render())
        self.display_handle.display(object)

    def update(self, change=None):
        if self.display_handle:
            self.display_handle.update(self.display_cls(self.render()))


class MarkdownItMixin:
    def render(self):
        return self.markdown_renderer.render(super().render())


@dataclass
class IPythonHtml(MarkdownItMixin, IPythonMarkdown):
    from IPython.display import HTML

    display_cls: type = HTML
    del HTML


@dataclass
class IPyWidgetsHtml(MarkdownItMixin, TemplateDisplay):
    widget = True
    from ipywidgets import HTML

    display_cls: object = HTML
    del HTML

    def display(self):
        from IPython.display import display

        if self.display_handle is None:
            self.display_handle = self.display_cls(self.render())
        display(self.display_handle)

    def update(self, change=None):
        if self.display_handle:
            self.display_handle.value = self.render()


class DisplaysManager(Extension):
    displays = Dict()
    prior = Dict()
    template_cls = Type(IPyWidgetsHtml, TemplateDisplay)
    markdown_renderer = Instance(MarkdownIt, args=())
    reactive = Bool(True)
    widgets = Dict()

    def weave_cell(self, body):
        template = self.shell.environment.from_string(body, None, IPythonTemplate)
        vars = find_undeclared_variables(self.shell.environment.parse(body))
        return self.template_cls(
            template=template, vars=vars, markdown_renderer=self.markdown_renderer
        )

    def get_value(self, key, raw=True):
        """get a value in a namespace that is potentially a widgets."""
        value = self.shell.user_ns.get(key)

        if raw:
            return value

        if is_widget(value):
            return value.value

        return value

    def get_id(self):
        return self.shell.kernel.get_parent().get("metadata", {}).get("cellId")

    def get_vars(self):
        data = defaultdict(list)
        for d in self.displays.values():
            if d.widget:
                for k in d.vars:
                    data[k].append(d)

        return data

    def link_widgets(self):
        displays = self.get_vars()
        for k, disp in displays.items():
            v = self.get_value(k)
            if is_widget(v):
                if v is not self.widgets.get(k):
                    self.widgets[k] = v
                    for d in disp:
                        v.observe(d.update)

    def post_run_cell(self, result):
        from IPython.display import display

        id = self.get_id()
        if result.error_in_exec or result.error_before_exec:
            pass  # don't do anything when there are errors
        elif NO_SHOW.match(result.info.raw_cell):
            pass
        elif CELL_MAGIC.match(result.info.raw_cell):
            pass
        else:
            disp = self.displays[id] = self.weave_cell(result.info.raw_cell)
            display(self.displays[id])
            return

        self.displays.pop(id, None)

    def pre_execute(self):
        metadata = self.shell.kernel.get_parent().get("metadata", {})
        for id in metadata.get("deletedCells", []):
            if id in self.displays:
                del self.displays[id]

        vars = set()
        for id, disp in self.displays.items():
            if disp.vars:
                vars.update(disp.vars)
        self.prior.update(zip(vars, map(self.get_value, vars)))

    def post_execute(self):
        changed = set()
        if self.reactive:
            self.link_widgets()

        for k, v in self.prior.items():
            y = self.get_value(k)
            if y is not v:
                changed.update({k})

        for id, disp in self.displays.items():
            changed.intersection(disp.vars) and disp.update()


def get_active_types(shell=None):
    """get the active types in the current IPython shell.
    we ignore latex, but i forget why."""
    shell = shell or get_ipython()
    if shell:
        object = list(shell.display_formatter.active_types)
        object.insert(object.index("text/html"), object.pop(object.index("text/latex")))
        return reversed(object)
    return []


def get_minified(x):
    from htmlmin import minify

    return minify(x, False, True, True, True, True, True, True)


def get_decoded(object):
    if isinstance(object, bytes):
        from base64 import b64encode

        object = b64encode(object).decode("utf-8")
    return object


def get_environment(reuse=True, _cache={}, **kwargs):
    # use this function to avoid repeat environment instantiation
    shell = get_ipython()
    if shell:
        try:
            return shell.environment
        except AttributeError:
            pass
    if reuse:
        if _cache:
            return _cache[True]
        _cache[True] = IPythonEnvironment(**kwargs)
        return _cache[True]
    return IPythonEnvironment(**kwargs)


def is_widget(object):
    """is an object a widget"""
    from sys import modules

    if "ipywidgets" in modules:
        from ipywidgets import Widget

        return isinstance(object, Widget)
    return False


def load_ipython_extension(shell):
    shell.add_traits(environment=Instance(IPythonEnvironment, default_value=get_environment()))
    shell.add_traits(displays_manager=Instance(Extension, allow_none=True))
    shell.displays_manager = DisplaysManager(shell=shell)
    shell.displays_manager.load_ipython_extension()


def unload_ipython_extension(shell):
    DisplaysManager(shell=shell).unload_ipython_extension()



"""pidgy literate computing frame"""


IS_IPY = bool(
    getattr(
        __import__("sys").modules.get("IPython", __import__("sys")),
        "get_ipython",
        lambda: None,
    )()
)


def load_ipython_extension(shell):
    from . import extras, tangle, weave

    shell.user_ns.setdefault("shell", shell)

    extras.load_ipython_extension(shell)
    tangle.load_ipython_extension(shell)
    weave.load_ipython_extension(shell)


def unload_ipython_extension(shell):
    from . import extras, tangle, weave

    extras.unload_ipython_extension(shell)
    tangle.unload_ipython_extension(shell)
    weave.unload_ipython_extension(shell)



from importnb import Notebook

from .extras import ReturnDisplay
from .markdown import Renderer
from .tangle import Python


class Markdown(Notebook):
    """an importnb extension for pidgy documents"""

    extensions = ".py.md", ".md", ".md.ipynb"
    tangle = Renderer(renderer_cls=Python)

    def get_data(self, path):
        if self.path.endswith(".md"):
            self.source = self.decode()
            return self.code(self.source)
        return super(Notebook, self).get_data(path)

    def code(self, str):
        return super().code(self.tangle.render("".join(str)))

    def visit(self, node):
        return super().visit(ReturnDisplay().visit(node))

    get_source = get_data = get_data



from pathlib import Path

from mkdocs_macros.plugin import MacrosPlugin


class PidgyPlugin(MacrosPlugin):
    def on_config(self, config):
        self.config.setdefault("modules", [])
        self.config["modules"].append(__name__)
        super().on_config(config)

    def on_page_read_source(self, page, config):
        with Path(config["docs_dir"], page.file.src_path).open() as file:
            first = next(file)
            if first.startswith("#!/usr/bin/env"):
                page.shebang = first
            return "".join(file)


def define_env(env):
    pass


def on_pre_page_macros(self):
    print("holler")
    if self.page.file.src_path.endswith((".md",)):
        from midgy.run import Markdown

        if getattr(self.page, "shebang", None):
            module = Markdown().load_code(self.page.markdown, self.page.file.src_path)
            self.variables.update(vars(module))



from ast import NodeTransformer
from importlib import import_module

from traitlets import Bool, CUnicode, Dict, HasTraits, Instance, List, default


def get_ipython():
    import IPython

    shell = IPython.get_ipython()
    if shell is None:
        shell = IPython.InteractiveShell()
        IPython.get_ipython = shell.get_ipython
    return shell


MAGICS = {"line", "line_cell", "cell"}
TRANSFORMS = {"cleanup_transforms"}


class Extension(HasTraits):
    from . import IS_IPY

    IS_IPY = True

    alias = CUnicode(allow_none=True)
    shell = Instance("IPython.InteractiveShell")
    enabled = Bool(True)

    def __init_subclass__(cls, alias=None, **kwargs):
        if alias:
            cls.alias = CUnicode(default=alias)
        super().__init_subclass__(**kwargs)

    def load_ipython_extension(self):
        # if self.register:
        #     # register a trait with a specific name on the current shell
        #     if not self.shell.has_trait(self.alias):
        #         self.shell.add_traits(**{self.alias: Instance(type(self))})

        #     # set the attribute on the shell
        #     setattr(self.shell, self.alias, self)

        # register any methods corresponding the shell events callbacks names
        for event in self.shell.events.callbacks:
            property = getattr(self, event, None)
            if property is not None:
                self.shell.events.register(event, property)

        vars = set(dir(self))

        for magic in set(MAGICS).intersection(vars):
            self.shell.register_magic_function(getattr(self, magic), magic, self.alias)

        if isinstance(self, NodeTransformer):
            self.shell.ast_transformers.append(self)

        for transform in TRANSFORMS.intersection(vars):
            getattr(self.shell.input_transformer_manager, transform).insert(
                0, getattr(self, transform)
            )

        return self

    def unload_ipython_extension(self):
        for event in self.shell.events.callbacks:
            property = getattr(self, event, None)
            if property is not None:
                try:
                    self.shell.events.unregister(event, property)
                except ValueError:
                    pass

        vars = set(dir(self))

        if isinstance(self, NodeTransformer):
            self.shell.ast_transformers = [x for x in self.shell.ast_transformers if x is not self]

        for transform in TRANSFORMS.intersection(vars):
            f = getattr(self, transform)
            ts = getattr(self.shell.input_transformer_manager, transform)
            setattr(
                self.shell.input_transformer_manager,
                transform,
                list(filter(f.__ne__, ts)),
            )

        return self



"""the extras extension adds features that improve interactive computing

1. exposes the `shell` variable for quick access to the interactive instance.
2. each execution the sys modules are populated in the namespace.
  
  this feature minimizes time between computing iterations by bringing more completion sooner.
  in an interactive computing session, the sys modules hold variable names not available to
  the author, but when available help continue flow between executions.

3. IPython display objects are added to the user namespace
4. top-level return statement become IPython.display expressions.
5. it should be easy to copy and paste json for beginners. 
     `true, false and null` are buultins.
"""

import builtins
from ast import Call, Expr, NodeTransformer, Tuple, parse
from pathlib import Path
from subprocess import CalledProcessError
from sys import modules

import IPython
from traitlets import CUnicode

from .pidgy import Extension

builtins.true, builtins.false, builtins.null = True, False, None


class SysModules(Extension):
    def pre_execute(self):
        """update the shell's user namespace to include imported modules"""
        for k in modules:
            if k:
                if k.startswith("_"):
                    continue
                if "." in k:
                    continue
                self.shell.user_ns.setdefault(k, modules[k])


class IPythonDisplays(Extension):
    def load_ipython_extension(self):
        """extract the display object from IPython"""
        from IPython import display
        from IPython.display import DisplayObject, IFrame, TextDisplayObject

        for k, v in vars(display).items():
            if isinstance(v, type) and issubclass(v, (DisplayObject, IFrame)):
                if v not in {DisplayObject, TextDisplayObject}:
                    if k[0].isupper():
                        self.shell.user_ns.setdefault(k, v)


class Shebang(Extension):
    alias = CUnicode("/usr/bin/env")

    def cleanup_transforms(self, lines):
        for i, line in enumerate(map(str.strip, lines)):
            if line:
                if line.startswith(("#!",)):
                    lines[i] = lines[0].replace("#!", "%%", 1)
                break
        return lines

    def cell(self, argv, body):
        import tempfile

        from IPython import get_ipython

        shell = get_ipython()
        with tempfile.NamedTemporaryFile(
            prefix=f"ipython-{shell.execution_count}", delete=False, suffix=".py"
        ) as file:
            file.write(body.encode())
        try:
            get_ipython().system(argv + " " + file.name)
        except CalledProcessError:
            pass
        finally:
            Path(file.name).unlink()


class ReturnDisplay(Extension, NodeTransformer):
    """transform a return node into an IPython display expression"""

    REPLACEMENT = parse(
        Extension.IS_IPY and '__import__("IPython").display.display' or "print",
        mode="eval",
    )

    def visit_FunctionDef(self, node):
        return node

    visit_AsyncFunctionDef = visit_FunctionDef

    def visit_Return(self, node):
        return Expr(
            Call(
                func=self.REPLACEMENT.body,
                args=node.value.elts if isinstance(node.value, Tuple) else [node.value],
                keywords=[],
            )
        )


def load_ipython_extension(shell: IPython.InteractiveShell):
    SysModules(shell=shell).load_ipython_extension()
    IPythonDisplays(shell=shell).load_ipython_extension()
    ReturnDisplay(shell=shell).load_ipython_extension()
    Shebang(shell=shell).load_ipython_extension()


def unload_ipython_extension(shell: IPython.InteractiveShell):
    SysModules(shell=shell).unload_ipython_extension()
    IPythonDisplays(shell=shell).unload_ipython_extension()
    ReturnDisplay(shell=shell).unload_ipython_extension()
    Shebang(shell=shell).unload_ipython_extension()



"""tangle.py provides the Markdown to Python translation model.
"""
from io import StringIO
from urllib.parse import urlparse

from midgy.python import Python


class IPython(Python):
    URL_PROTOCOLS = "file", "http", "https"
    VALID_URL_LIST_TOKENS = {
        "paragraph_open",
        "paragraph_close",
        "bullet_list_open",
        "bullet_list_close",
        "list_item_open",
        "list_item_close",
    }

    def _get_url_list(self, tokens):
        urls = []
        for token in tokens:
            if token.type in self.VALID_URL_LIST_TOKENS:
                continue
            elif token.type == "inline":
                for line in StringIO(token.content):
                    parsed = urlparse(line)
                    if parsed.netloc in self.URL_PROTOCOLS:
                        urls.append(line.strip())
                        continue
                    break
                else:
                    continue
                break
            break
        else:
            return urls


def load_ipython_extension(shell):
    from traitlets import Instance

    def tangle(line, cell):
        print(shell.tangle.render(cell))

    def parse(line, cell):
        print(shell.tangle.parse(cell))

    shell.add_traits(tangle=Instance(IPython, ()))
    shell.input_transformer_manager.cleanup_transforms.insert(0, shell.tangle.render_lines)
    shell.register_magic_function(tangle, "cell")
    shell.register_magic_function(parse, "cell")


def unload_ipython_extension(shell):
    if shell.has_trait("tangle"):
        shell.input_transformer_manager.cleanup_transforms = list(
            filter(
                shell.tangle.render_lines.__ne__,
                shell.input_transformer_manager.cleanup_transforms,
            )
        )
